from quart import request
import functools


class EasyRequest:
    """
    A helper class that make request handling easier, especially for beginners.
    All attributes are assessable via the same syntax, while with Flask.request
    or quart.request, you will have slightly different syntax when retrieving
    different request attributes.

    ..hint:
        You only need to await attributes that need calculation, for example
        evaluating the request body, like `.json` or `.form`.

    """

    params = ...
    headers = ...
    __quart_request = ...

    def load(self):
        self.__quart_request = request
        self.params = dict(self.__quart_request.args)
        self.headers = dict(self.__quart_request.headers)
        return self

    def __call__(self, f):
        print("decorating")

        @functools.wraps(f)
        async def decorated_function(*args2, **kwargs2):
            f.__globals__[self.__class__.__name__] = self.__class__().load()
            return await f(*args2, **kwargs2)

        return decorated_function

    @property
    async def form(self):
        return dict(await self.__quart_request.form)

    @property
    async def json(self):
        json = await self.__quart_request.get_json()
        return {} if json is None else json


if __name__ == '__main__':
    from Aeros import WebServer, render_template, jsonify
    from time import strftime

    app = WebServer(__name__, host="0.0.0.0", port=80)  # init backend & web server
    app.template_folder = "./frontend/build/"
    app.static_folder = "./frontend/build/static/"


    @app.route("/")
    async def home():
        return await render_template("index.html")


    @app.route("/time/")
    @EasyRequest()
    async def current_time():
        print("Headers:", EasyRequest.headers)
        print("Params :", EasyRequest.params)
        print("Form   :", await EasyRequest.form)
        print("JSON   :", await EasyRequest.json)
        return jsonify({"timestamp": strftime("%H:%M:%S")})


    app.run_server()  # run web server in this thread (endless mode)



from .WebServer import *
from .misc import *

from .threading import AdvancedThread
from .compression import Compression
from .caching import (
    SimpleCache,
    Cache,
    FilesystemCache,
    RedisCache
)



from .patches.flask_caching import Cache


class SimpleCache(Cache):
    def __init__(self, *args, **kwargs):
        Cache.__init__(self, *args, **kwargs)
        self.config["CACHE_TYPE"] = "simple"


class FilesystemCache(Cache):
    def __init__(self, directory: str, *args, **kwargs):
        Cache.__init__(self, *args, **kwargs)
        self.config["CACHE_TYPE"] = "filesystem"
        self.config["CACHE_DIR"] = directory


class RedisCache(Cache):
    def __init__(self, host: str, port: int, password: str = "", db: int = 0, *args, **kwargs):
        Cache.__init__(self, *args, **kwargs)
        self.config["CACHE_TYPE"] = "redis"
        self.config["CACHE_REDIS_HOST"] = host
        self.config["CACHE_REDIS_PORT"] = port
        self.config["CACHE_REDIS_PASSWORD"] = password
        self.config["CACHE_REDIS_DB"] = db



from flask_caching import Cache as OriginalCache
from flask_caching import *


class Cache(OriginalCache):
    def __init__(self, timeout: int = 60 * 60, threshold: int = 100, *args, **kwargs):
        OriginalCache.__init__(self, config={}, *args, **kwargs)
        self.config["CACHE_TYPE"] = "null"
        self.config["CACHE_THRESHOLD"] = threshold
        self.config["CACHE_DEFAULT_TIMEOUT"] = timeout

    def cached(
            self,
            timeout=None,
            key_prefix="view/%s",
            unless=None,
            forced_update=None,
            response_filter=None,
            query_string=False,
            hash_method=hashlib.md5,
            cache_none=False,
    ):
        """Decorator. Use this to cache a function. By default the cache key
        is `view/request.path`. You are able to use this decorator with any
        function by changing the `key_prefix`. If the token `%s` is located
        within the `key_prefix` then it will replace that with `request.path`

        Example::

            # An example view function
            @cache.cached(timeout=50)
            def big_foo():
                return big_bar_calc()

            # An example misc function to cache.
            @cache.cached(key_prefix='MyCachedList')
            def get_list():
                return [random.randrange(0, 1) for i in range(50000)]

            my_list = get_list()

        .. note::

            You MUST have a request context to actually called any functions
            that are cached.

        .. versionadded:: 0.4
            The returned decorated function now has three function attributes
            assigned to it. These attributes are readable/writable.

                **uncached**
                    The original undecorated function

                **cache_timeout**
                    The cache timeout value for this function. For a
                    custom value to take affect, this must be set before the
                    function is called.

                **make_cache_key**
                    A function used in generating the cache_key used.

        :param timeout: Default None. If set to an integer, will cache for that
                        amount of time. Unit of time is in seconds.

        :param key_prefix: Default 'view/%(request.path)s'. Beginning key to .
                           use for the cache key. `request.path` will be the
                           actual request path, or in cases where the
                           `make_cache_key`-function is called from other
                           views it will be the expected URL for the view
                           as generated by Flask's `url_for()`.

                           .. versionadded:: 0.3.4
                               Can optionally be a callable which takes
                               no arguments but returns a string that will
                               be used as the cache_key.

        :param unless: Default None. Cache will *always* execute the caching
                       facilities unless this callable is true.
                       This will bypass the caching entirely.

        :param forced_update: Default None. If this callable is true,
                              cache value will be updated regardless cache
                              is expired or not. Useful for background
                              renewal of cached functions.

        :param response_filter: Default None. If not None, the callable is
                                invoked after the cached funtion evaluation,
                                and is given one arguement, the response
                                content. If the callable returns False, the
                                content will not be cached. Useful to prevent
                                caching of code 500 responses.

        :param query_string: Default False. When True, the cache key
                             used will be the result of hashing the
                             ordered query string parameters. This
                             avoids creating different caches for
                             the same query just because the parameters
                             were passed in a different order. See
                             _make_cache_key_query_string() for more
                             details.

        :param hash_method: Default hashlib.md5. The hash method used to
                            generate the keys for cached results.
        :param cache_none: Default False. If set to True, add a key exists
                           check when cache.get returns None. This will likely
                           lead to wrongly returned None values in concurrent
                           situations and is not recommended to use.

        """

        def decorator(f):
            @functools.wraps(f)
            async def decorated_function(*args, **kwargs):
                #: Bypass the cache entirely.
                if self._bypass_cache(unless, f, *args, **kwargs):
                    return await f(*args, **kwargs)

                try:

                    cache_key = "{}{}{}".format(f.__name__, args, kwargs)

                    if (
                            callable(forced_update)
                            and (
                            forced_update(*args, **kwargs)
                            if wants_args(forced_update)
                            else forced_update()
                    )
                            is True
                    ):
                        rv = None
                        found = False
                    else:
                        rv = self.cache.get(cache_key)
                        found = True

                        # If the value returned by cache.get() is None, it
                        # might be because the key is not found in the cache
                        # or because the cached value is actually None
                        if rv is None:
                            # If we're sure we don't need to cache None values
                            # (cache_none=False), don't bother checking for
                            # key existence, as it can lead to false positives
                            # if a concurrent call already cached the
                            # key between steps. This would cause us to
                            # return None when we shouldn't
                            if not cache_none:
                                found = False
                            else:
                                found = self.cache.has(cache_key)
                except Exception:
                    if self.app.debug:
                        raise
                    logger.exception("Exception possibly due to cache backend.")
                    return await f(*args, **kwargs)

                if not found:
                    rv = await f(*args, **kwargs)

                    if response_filter is None or response_filter(rv):
                        try:
                            self.cache.set(
                                cache_key,
                                rv,
                                timeout=decorated_function.cache_timeout,
                            )
                        except Exception:
                            if self.app.debug:
                                raise
                            logger.exception(
                                "Exception possibly due to cache backend."
                            )
                return rv

            def make_cache_key(*args, **kwargs):
                # Convert non-keyword arguments (which is the way
                # `make_cache_key` expects them) to keyword arguments
                # (the way `url_for` expects them)
                argspec_args = inspect.getfullargspec(f).args

                for arg_name, arg in zip(argspec_args, args):
                    kwargs[arg_name] = arg

                return _make_cache_key(args, kwargs, use_request=False)

            def _make_cache_key_query_string():
                """Create consistent keys for query string arguments.

                Produces the same cache key regardless of argument order, e.g.,
                both `?limit=10&offset=20` and `?offset=20&limit=10` will
                always produce the same exact cache key.
                """

                # Create a tuple of (key, value) pairs, where the key is the
                # argument name and the value is its respective value. Order
                # this tuple by key. Doing this ensures the cache key created
                # is always the same for query string args whose keys/values
                # are the same, regardless of the order in which they are
                # provided.
                args_as_sorted_tuple = tuple(
                    sorted((pair for pair in request.args.items(multi=True)))
                )
                # ... now hash the sorted (key, value) tuple so it can be
                # used as a key for cache. Turn them into bytes so that the
                # hash function will accept them
                args_as_bytes = str(args_as_sorted_tuple).encode()
                hashed_args = str(hash_method(args_as_bytes).hexdigest())
                cache_key = request.path + hashed_args
                return cache_key

            def _make_cache_key(args, kwargs, use_request):
                if callable(key_prefix):
                    cache_key = key_prefix()
                elif "%s" in key_prefix:
                    if use_request:
                        cache_key = key_prefix % request.path
                    else:
                        cache_key = key_prefix % url_for(f.__name__, **kwargs)
                else:
                    cache_key = key_prefix

                return cache_key

            decorated_function.uncached = f
            decorated_function.cache_timeout = timeout
            decorated_function.make_cache_key = make_cache_key

            return decorated_function

        return decorator



import platform
import random
import time
from multiprocessing import Event
from Aeros.threading import AdvancedThread
from hypercorn.config import Config
from hypercorn.utils import write_pid_file

from .worker import asyncio_worker


def run(app, config):
    if config.pid_path is not None:
        write_pid_file(config.pid_path)

    if config.worker_class != "asyncio":
        raise ValueError(f"No worker of class {config.worker_class} exists")

    if config.workers == 1:
        asyncio_worker(app, config)
    else:
        run_multiple(app, config, asyncio_worker)


def run_multiple(app, config: Config, worker_func: asyncio_worker) -> None:
    if config.use_reloader:
        raise RuntimeError("Reloader can only be used with a single worker")

    sockets = config.create_sockets()

    processes = []

    shutdown_event = Event()

    for _ in range(config.workers):
        process = AdvancedThread(
            target=worker_func,
            kwargs={"app": app, "config": config, "shutdown_event": shutdown_event, "sockets": sockets},
        )
        process.daemon = True
        process.start()
        processes.append(process)
        if platform.system() == "Windows":
            time.sleep(0.1 * random.random())

    for process in processes:
        process.join()
    for process in processes:
        process.stop()

    for sock in sockets.secure_sockets:
        sock.close()
    for sock in sockets.insecure_sockets:
        sock.close()



from hypercorn.asyncio.run import *
from hypercorn.asyncio.run import _run, _share_socket, _windows_signal_support


async def worker_serve(
        app: ASGIFramework,
        config: Config,
        *,
        sockets: Optional[Sockets] = None,
        shutdown_trigger: Optional[Callable[..., Awaitable[None]]] = None,
) -> None:
    config.set_statsd_logger_class(StatsdLogger)

    lifespan = Lifespan(app, config)
    lifespan_task = asyncio.ensure_future(lifespan.handle_lifespan())

    await lifespan.wait_for_startup()
    if lifespan_task.done():
        exception = lifespan_task.exception()
        if exception is not None:
            raise exception

    if sockets is None:
        sockets = config.create_sockets()

    loop = asyncio.get_event_loop()
    tasks = []
    if platform.system() == "Windows":
        tasks.append(loop.create_task(_windows_signal_support()))

    if shutdown_trigger is None:
        shutdown_trigger = asyncio.Future
    tasks.append(loop.create_task(raise_shutdown(shutdown_trigger)))

    if config.use_reloader:
        tasks.append(loop.create_task(observe_changes(asyncio.sleep)))

    ssl_handshake_timeout = None
    if config.ssl_enabled:
        ssl_context = config.create_ssl_context()
        ssl_handshake_timeout = config.ssl_handshake_timeout

    async def _server_callback(reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        try:
            await TCPServer(app, loop, config, reader, writer)
        except ConnectionAbortedError:
            pass

    servers = []
    for sock in sockets.secure_sockets:
        if config.workers > 1 and platform.system() == "Windows":
            sock = _share_socket(sock)

        servers.append(
            await asyncio.start_server(
                _server_callback,
                backlog=config.backlog,
                loop=loop,
                ssl=ssl_context,
                sock=sock,
                ssl_handshake_timeout=ssl_handshake_timeout,
            )
        )
        bind = repr_socket_addr(sock.family, sock.getsockname())
        await config.log.info(f"Running on {bind} over https (CTRL + C to quit)")

    for sock in sockets.insecure_sockets:
        if config.workers > 1 and platform.system() == "Windows":
            sock = _share_socket(sock)

        servers.append(
            await asyncio.start_server(
                _server_callback, backlog=config.backlog, loop=loop, sock=sock
            )
        )
        bind = repr_socket_addr(sock.family, sock.getsockname())
        await config.log.info(f"Running on {bind} over http (CTRL + C to quit)")

    for sock in sockets.quic_sockets:
        if config.workers > 1 and platform.system() == "Windows":
            sock = _share_socket(sock)

        await loop.create_datagram_endpoint(lambda: UDPServer(app, loop, config), sock=sock)
        bind = repr_socket_addr(sock.family, sock.getsockname())
        await config.log.info(f"Running on {bind} over quic (CTRL + C to quit)")

    reload_ = False
    try:
        gathered_tasks = asyncio.gather(*tasks)
        await gathered_tasks

    except MustReloadException:
        reload_ = True
    except (Shutdown, KeyboardInterrupt):
        pass
    finally:
        for server in servers:
            server.close()
            await server.wait_closed()

        # Retrieve the Gathered Tasks Cancelled Exception, to
        # prevent a warning that this hasn't been done.
        gathered_tasks.exception()

        await lifespan.wait_for_shutdown()
        lifespan_task.cancel()
        await lifespan_task

    if reload_:
        restart()


def asyncio_worker(app, config: Config, sockets: Optional[Sockets] = None, shutdown_event: Optional[EventType] = None) -> None:
    shutdown_trigger = None
    if shutdown_event is not None:
        shutdown_trigger = partial(check_multiprocess_shutdown_event, shutdown_event, asyncio.sleep)

    if config.workers > 1 and platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore

    _run(
        partial(worker_serve, app, config, sockets=sockets),
        debug=config.debug,
        shutdown_trigger=shutdown_trigger,
    )



"""
Main web server instance
"""

import functools
import argparse
import warnings
import inspect
import ssl
from typing import Union, Dict
from .patches.quart.app import Quart
import hashlib

from .patches.hypercorn import run,Config
from .caching import Cache
from .compression import Compression


def make_config_from_hypercorn_args(hypercorn_string: str, config: Config = Config()) -> Config:
    """ Overrides a given config's items if they are specified in the hypercorn args string """

    def _convert_verify_mode(value: str) -> ssl.VerifyMode:
        try:
            return ssl.VerifyMode[value]
        except KeyError:
            raise argparse.ArgumentTypeError("Not a valid verify mode")

    sentinel = object()
    parser = argparse.ArgumentParser()
    parser.add_argument("--access-log", default=sentinel)
    parser.add_argument("--access-logfile", default=sentinel)
    parser.add_argument("--access-logformat", default=sentinel)
    parser.add_argument("--backlog", type=int, default=sentinel)
    parser.add_argument("-b", "--bind", dest="binds", default=[], action="append", )
    parser.add_argument("--ca-certs", default=sentinel)
    parser.add_argument("--certfile", default=sentinel)
    parser.add_argument("--cert-reqs", type=int, default=sentinel)
    parser.add_argument("--ciphers", default=sentinel)
    parser.add_argument("-c", "--config", default=None)
    parser.add_argument("--debug", action="store_true", default=sentinel)
    parser.add_argument("--error-log", default=sentinel)
    parser.add_argument("--error-logfile", "--log-file", dest="error_logfile", default=sentinel)
    parser.add_argument("-g", "--group", default=sentinel, type=int)
    parser.add_argument("-k", "--worker-class", dest="worker_class", default=sentinel)
    parser.add_argument("--keep-alive", default=sentinel, type=int)
    parser.add_argument("--keyfile", default=sentinel)
    parser.add_argument("--insecure-bind", dest="insecure_binds", default=[], action="append")
    parser.add_argument("--log-config", default=sentinel)
    parser.add_argument("--log-level", default="info")
    parser.add_argument("-p", "--pid", default=sentinel)
    parser.add_argument("--quic-bind", dest="quic_binds", default=[], action="append")
    parser.add_argument("--reload", action="store_true", default=sentinel)
    parser.add_argument("--root-path", default=sentinel)
    parser.add_argument("--statsd-host", default=sentinel)
    parser.add_argument("--statsd-prefix", default="")
    parser.add_argument("-m", "--umask", default=sentinel, type=int)
    parser.add_argument("-u", "--user", default=sentinel, type=int)
    parser.add_argument("-w", "--workers", dest="workers", default=sentinel, type=int)
    parser.add_argument("--verify-mode", type=_convert_verify_mode, default=sentinel)

    args = hypercorn_string.split(" ").remove("") if "" in hypercorn_string.split(" ") else hypercorn_string.split(" ")
    args = parser.parse_args(args)

    config.loglevel = args.log_level
    if args.access_logformat is not sentinel:
        config.access_log_format = args.access_logformat
    if args.access_log is not sentinel:
        warnings.warn("The --access-log argument is deprecated, use `--access-logfile` instead", DeprecationWarning, )
        config.accesslog = args.access_log
    if args.access_logfile is not sentinel:
        config.accesslog = args.access_logfile
    if args.backlog is not sentinel:
        config.backlog = args.backlog
    if args.ca_certs is not sentinel:
        config.ca_certs = args.ca_certs
    if args.certfile is not sentinel:
        config.certfile = args.certfile
    if args.cert_reqs is not sentinel:
        config.cert_reqs = args.cert_reqs
    if args.ciphers is not sentinel:
        config.ciphers = args.ciphers
    if args.debug is not sentinel:
        config.debug = args.debug
    if args.error_log is not sentinel:
        warnings.warn("The --error-log argument is deprecated, use `--error-logfile` instead", DeprecationWarning, )
        config.errorlog = args.error_log
    if args.error_logfile is not sentinel:
        config.errorlog = args.error_logfile
    if args.group is not sentinel:
        config.group = args.group
    if args.keep_alive is not sentinel:
        config.keep_alive_timeout = args.keep_alive
    if args.keyfile is not sentinel:
        config.keyfile = args.keyfile
    if args.log_config is not sentinel:
        config.logconfig = args.log_config
    if args.pid is not sentinel:
        config.pid_path = args.pid
    if args.root_path is not sentinel:
        config.root_path = args.root_path
    if args.reload is not sentinel:
        config.use_reloader = args.reload
    if args.statsd_host is not sentinel:
        config.statsd_host = args.statsd_host
    if args.statsd_prefix is not sentinel:
        config.statsd_prefix = args.statsd_prefix
    if args.umask is not sentinel:
        config.umask = args.umask
    if args.user is not sentinel:
        config.user = args.user
    if args.worker_class is not sentinel:
        config.worker_class = args.worker_class
    if args.workers is not sentinel:
        config.workers = args.workers
    if len(args.binds) > 0:
        config.bind = args.binds
    if len(args.insecure_binds) > 0:
        config.insecure_bind = args.insecure_binds
    if len(args.quic_binds) > 0:
        config.quic_bind = args.quic_binds

    return config


class WebServer(Quart):
    """ This is the main server class which extends a standard Flask class by a bunch of features and major
    performance improvements. It extends the Quart class, which by itself is already an enhanced version of
    the Flask class. This class however allows production-grade  deployment using the hypercorn WSGI server
    as production server. But instead of calling the hypercorn command via the console, it can be started
    directly from the Python code itself, making it easier to integrate in higher-level scripts and
    applications without calling os.system() od subprocess.Popen(). """

    def __init__(self, import_name: str, host: str = "0.0.0.0", port: int = 80, include_server_header: bool = True,
                 hypercorn_arg_string: str = "", worker_threads: int = 1, logging_level: Union[int, str] = "INFO",
                 cache: Cache = Cache(), compression: Compression = Compression(level=2, min_size=10),
                 global_headers: Dict[str, str] = None,
                 *args, **kwargs):

        super().__init__(import_name, *args, **kwargs)

        self.logger.setLevel(logging_level)
        self._host, self._port = host, port
        self._global_headers = global_headers
        self._hypercorn_arg_string = hypercorn_arg_string
        self._worker_threads = worker_threads
        self._include_server_header = include_server_header

        self._cache = cache
        self._compression = compression

    def _get_own_instance_path(self):
        """ DEPRECATED!
         Since hypercorn needs the application's file and global variable name, an instance needs to know
         it's own origin file and name. But since this class is not defined in the same file as it is called
         or defined from, this method searches for the correct module/file and evaluates it's instance name. """
        self.logger.warning("_get_own_instance_path() is deprecated!")
        for i in range(10):
            try:
                frame = inspect.stack()[i][0].f_globals.items()
                for key, val in frame:
                    try:
                        if val == self:
                            return f"{dict(frame)['__name__']}:{key}"
                    except (RuntimeError, AttributeError, KeyError):
                        pass
            except:
                pass
        raise Exception("QuartServer() is unable to find own instance file:name")

    def cache(self, timeout=None, key_prefix="view/%s", unless=None, forced_update=None,
              response_filter=None, query_string=False, hash_method=hashlib.md5, cache_none=False, ):
        """ A simple wrapper that forwards cached() decorator to the internal
        Cache() instance. May be used as the normal @cache.cached() decorator. """

        def decorator(f):
            @functools.wraps(f)
            @self._cache.cached(timeout=timeout, key_prefix=key_prefix, unless=unless, forced_update=forced_update,
                                response_filter=response_filter, query_string=query_string, hash_method=hash_method, cache_none=cache_none)
            async def decorated_function(*args2, **kwargs2):
                x = await f(*args2, **kwargs2)
                return x

            return decorated_function

        return decorator

    def run_server(self) -> None:
        """ Generates the necessary config and runs the server instance. """

        config = Config(self._global_headers)

        config.bind = [f'{self._host}:{self._port}']
        config.workers = self._worker_threads
        config.include_server_header = self._include_server_header

        # override config items if specified in hypercorn arguments
        config = make_config_from_hypercorn_args(self._hypercorn_arg_string, config=config)

        # Initialize extra features just in case the user replaced them with their own instances
        self._cache.init_app(self)
        if type(self._compression == Compression):
            self._compression.init_app(self)

        run(self, config)



from hypercorn.config import Config as OriginalConfig
from hypercorn.config import format_date_time, List, Tuple, time, Dict


class Config(OriginalConfig):
    def __init__(self, custom_headers: Dict[str, str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__custom_headers = custom_headers if custom_headers else {}

    def response_headers(self, protocol: str) -> List[Tuple[bytes, bytes]]:
        """ This function is patched to include custom headers, which will
         be sent in every response. For example to send a custom "server"
         header or CORS headers all the time. """

        headers = [(b"date", format_date_time(time()).encode("ascii"))]
        if self.include_server_header:
            headers.append((b"server", f"hypercorn-{protocol}".encode("ascii")))

        for alt_svc_header in self.alt_svc_headers:
            headers.append((b"alt-svc", alt_svc_header.encode()))
        if len(self.alt_svc_headers) == 0 and self._quic_bind:
            from aioquic.h3.connection import H3_ALPN

            for version in H3_ALPN:
                for bind in self._quic_bind:
                    port = int(bind.split(":")[-1])
                    headers.append((b"alt-svc", b'%s=":%d"; ma=3600' % (version.encode(), port)))

        for header, value in self.__custom_headers.items():
            headers.append((header.encode("ascii"), value.encode("ascii")))

        return headers
